(delcare
 (unit fume))
(module
 fume
 (define-transformer)
(import scheme chicken)

(define-syntax extract
  ;; # Petrofsky Extraction
  ;;
  ;; From: Al Petrofsky
  ;; Subject: How to write seemingly unhygienic macros using syntax-rules
  ;; Date: 2001-11-19 01:23:33 PST
  ;;
  ;; Extended to extract multiple identifier by Oleg Kiselyov.
  ;; Rewritten as single annonymous syntax-rules by Jörg F. Wittenberger.
  ;;
  ;; Extract several colored identifiers from a form
  ;;    SYMB-L BODY CONT
  ;; where
  ;;
  ;; SYMB-L is the list of symbols SYMB to extract
  ;;
  ;; BODY is a form that may contain an occurence of an identifiers
  ;; that refers to the same binding occurrence as SYMB, perhaps with
  ;; a different color.
  ;;
  ;; CONT is a form of the shape (K-HEAD K-IDL . K-ARGS) where K-IDL
  ;; are K-ARGS are S-expressions representing lists or the empty
  ;; list.
  ;;
  ;; The macro expands into
  ;;
  ;;   (K-HEAD (extr-id-l . K-IDL) . K-ARGS)
  ;;
  ;; where extr-id-l is the list of extracted colored identifiers.
  ;;
  ;; License: public domain - reuse as at your discretion.
  (syntax-rules ()
    ((_ symb-l body cont)
     (let-syntax
	 (
	  ;; Extract one colored identifier from a form
	  ;;    extract SYMB BODY CONT
	  ;;
	  ;; The extract macro expands into
	  ;;   (K-HEAD (extr-id . K-IDL) . K-ARGS)
	  ;; where extr-id is the extracted colored identifier. If
	  ;; symbol SYMB does not occur in BODY at all, extr-id is
	  ;; identical to SYMB.
	 (extract
	  (syntax-rules ()
	    ((_ symb body_1 cont_1)
	     (letrec-syntax
		 ((tr
		   (syntax-rules (symb)
		     ((_ x symb tail (cont-head symb-l_1 . cont-args))
		      (cont-head (x . symb-l_1) . cont-args)) ; symb has occurred
		     ((_ d (x . y) tail cont_2)   ; if body is a composite form,
		      (tr x x (y . tail) cont_2)) ; look inside
		     ((_ d1 d2 () (cont-head  symb-l_1 . cont-args))
		      (cont-head (symb . symb-l_1) . cont-args)) ; symb does not occur
		     ((_ d1 d2 (x . y) cont_2)
		      (tr x x y cont_2)))))
	       (tr body_1 body_1 () cont_1))))))
       (let-syntax
	   ((_extract*
	     (syntax-rules ()
	       ((_ (symb) body_1 cont_1) ; only one symbol: use extract to do the job
		(extract symb body_1 cont_1))
	       ((_ _symbs _body _cont_1)
		(letrec-syntax
		    ((ex-aux			; extract symbol-by-symbol
		      (syntax-rules ()
			((_ found-symbs () body_1 cont_2)
			 (reverse () found-symbs cont_2))
			((_ found-symbs (symb . symb-others) body_1 cont_2)
			 (extract symb body_1
				  (ex-aux found-symbs symb-others body_1 cont_2)))
			))
		     (reverse	       ; reverse the list of extracted symbols
		      (syntax-rules ()	; to match the order of SYMB-L
			((_ res () (cont-head () . cont-args))
			 (cont-head res . cont-args))
			((_ res (x . tail) cont_2)
			 (reverse (x . res) tail cont_2)))))
		  (ex-aux () _symbs _body _cont_1))))))
	 (_extract* symb-l body cont))))))

(define-syntax let-keyword-syntax
  (syntax-rules ()
    ((_ () body) body)
    ((_
      ((keyword-syntax-name (positional-operator (arg-name . arg-def) ...)) . more)
       body ...)
     (letrec-syntax
	 ((keyword-syntax-name
	   (syntax-rules ()
	     ((_ . kw-val-pairs)
	      (letrec-syntax
		  ((find 
		    (syntax-rules (arg-name ...)
		      ((find k-args (arg-name . default) arg-name
			     val . others)	   ; found arg-name among kw-val-pairs
		       (next val . k-args)) ...
		       ((find k-args key arg-no-match-name val . others)
			(find k-args key . others))
		       ((find k-args (arg-name default)) ; default must be here
			(next default . k-args)) ...
			))
		   (next			; pack the continuation to find
		    (syntax-rules ()
		      ((next val vals key . keys)
		       (find ((val . vals) . keys) key . kw-val-pairs))
		      ((next val vals)	; processed all arg-descriptors
		       (rev-apply (val) vals))))
		   (rev-apply
		    (syntax-rules ()
		      ((rev-apply form (x . xs))
		       (rev-apply (x . form) xs))
		      ((rev-apply form ()) form))))
		(next positional-operator () 
		      (arg-name . arg-def) ...)))))
	  )
       (let-keyword-syntax more (begin body ...))))))

(define-syntax lambda-transformer
  (syntax-rules ()
    ((_ errloc body)
     (begin
       ;; (: name :render-renderer:)
       (lambda (place			; only for Askemos
		message			; only for Askemos
		root-node0
		variables		; xsl binding environment for
					; parameters and variables
		namespaces0
		ancestors0		; ancestor context
		self-node		; current transformer input
		nl			; current sosofo
		mode-choice0		; 1-ari proc, selects
					; continuation mode, #f
					; selects default mode, symbol
					; selects named mode, string
					; selects named template
		proc-chain		; next processing instruction (to
					; be remove)
		)
	 (let-syntax
	     ((rewrite
	       (syntax-rules ()
		 ((_ (*place
		      *message
		      *root-node
		      *environment
		      *namespaces
		      *ancestors
		      *$
		      *$<
		      *@?
		      *@!
		      *transform
		      *lambda-transformer
		      *apply-transformer
		      *current-node
		      *sosofo
		      *mode-choice
		      *%%proc-chain
		      )
		     body_)
		  (letrec-syntax
		      (
		       (*place (syntax-rules ($$)
				 ((_ $$) (place 'get 'id))
				 ((_) place)))
		       (*message (syntax-rules (format-location)
				   ((_) message)
				   ((_ format-location mode)
				    ((if (eq? mode 'get) read-locator write-locator)
				     (message 'location-format)
				     (message 'location)))
				   ((_ format-location mode tail)
				    ((if (eq? mode 'get) read-locator write-locator)
				     (message 'location-format)
				     tail))
				   ))
		       (proc-chlds
			(syntax-rules ()
			  ((_ root-node_ variables_ namespaces_ ancestors_
			      self-node_ nl_ mode-choice_ proc-chain_)
			   (proc-chain_
			    place message root-node_ variables_ namespaces_ ancestors_
			    self-node_ nl_ mode-choice_))))
		       (apply-transformer_
			(syntax-rules ()
			  ((_ transformer_ root-node_ variables_ namespaces_ ancestors_
			      self-node_ nl_ mode-choice_ proc-chain_)
			   (transformer_
			    place message root-node_ variables_ namespaces_ ancestors_
			    self-node_ nl_ mode-choice_ proc-chain_))))
		       (*mode-choice (syntax-rules ()
				       ((_ key) (mode-choice0 key))))
		       (*lambda-transformer
			(syntax-rules ()
			  ((_ errloc_1 body_1)
			   (let ()
			     (define-transformer local errloc_1 body_1)
			     local))
			  ((_ body_1)
			   (let ()
			     (*lambda-transformer "local-transformer" body_1)))))
		       (*$ (syntax-rules ()
			     ((_ name)
			      (fetch-xsl-variable/string name (xsl-environment-variables variables) errloc))))
		       (*$< (syntax-rules ()
			     ((_ name)
			      (fetch-xsl-variable/symbol name (xsl-parameters variables) errloc))))
		       (*@? (syntax-rules ()
			      ((_ name) (attribute-string name nl))))
		       (*@! (syntax-rules ()
			      ((_ name) (required-attribute-string name nl errloc))))
		       
		       (*ancestors (syntax-rules () ((_) ancestors0)))
		       (*namespaces (syntax-rules () ((_) namespaces0)))
		       (*environment (syntax-rules () ((_) variables)))
		       (*root-node (syntax-rules () ((_) root-node0)))
		       (*current-node (syntax-rules () ((_) self-node)))
		       (*sosofo (syntax-rules () ((_) nl)))
		       (*%%proc-chain (syntax-rules () ((_) proc-chain)))
		       )
		    (let-keyword-syntax
		     ((*transform
		       (proc-chlds
			(root-node: root-node0)
			(variables: variables)
			(namespaces: namespaces0)
			(ancestors: ancestors0)
			(current-node: self-node)
			(sosofos: nl)
			(continue: mode-choice0)
			(process: proc-chain)))
		      (*apply-transformer
		       (apply-transformer_
			(use:)
			(root-node: root-node0)
			(variables: variables)
			(namespaces: namespaces0)
			(ancestors: ancestors0)
			(current-node: self-node)
			(sosofos: nl)
			(continue: mode-choice0)
			(process: proc-chain)))
		      )
		     body_))))))
	   (extract
	    (
	     current-place
	     current-message
	     root-node
	     environment
	     namespaces
	     ancestors
	     $				; resolve variable to value
	     $<
	     @?				; optional attribute as string or #f
	     @!				; required attribute as string
	     transform			; process children
	     cc/lambda-transformer
	     apply-transformer
	     current-node
	     sosofo
	     mode-choice
	     %%proc-chain		; current transformation
	     )
	    body
	    (rewrite () body))))))))

(define-syntax define-transformer
  (syntax-rules ()
    ((_ transformer errloc body)
     (begin
       (define transformer (lambda-transformer errloc body))))))



)
